@@ -1,14 +1,58 @@
 """The base Simon models"""
 
-from .connection import get_database
+import warnings
+
+from .connection import get_database, pymongo_supports_mongoclient
 from .exceptions import MultipleDocumentsFound, NoDocumentFound
 from .query import Q, QuerySet
 from .utils import (current_datetime, get_nested_key, guarantee_object_id,
-                    map_fields, remove_nested_key, update_nested_keys)
+                    is_atomic, map_fields, remove_nested_key,
+                    update_nested_keys)
 
 __all__ = ('Model',)
 
 
+# In version 2.4, PyMongo introduced MongoClient as a
+# replcaement for Connection. Part of the new class is that the
+# safe argument was deprecated in favor of the w argument.
+# Whereas safe was a boolean, w is an integer representing the
+# number of servers in a replica set that must receive the
+# update before the write is considered successful.
+
+def _set_write_concern_as_safe(options, force_safe):
+    safe = options.pop('safe', None)
+    w = options.pop('w', None)
+
+    if force_safe:
+        safe = True
+    else:
+        if safe is None and w:
+            safe = True
+        else:
+            safe = safe or False
+    options['safe'] = safe
+
+
+def _set_write_concern_as_w(options, force_safe):
+    safe = options.pop('safe', None)
+    w = options.pop('w', None)
+
+    if force_safe:
+        w = int(force_safe)  # use int() in case force_safe is True
+    else:
+        if w is None and safe:
+            w = 1
+        else:
+            w = w or 0
+    options['w'] = w
+
+
+if pymongo_supports_mongoclient:
+    _set_write_concern = _set_write_concern_as_w
+else:
+    _set_write_concern = _set_write_concern_as_safe
+
+
 class Meta(object):
     """Custom options for a :class:`~simon.Model`.
 
@@ -225,14 +269,6 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
-        if cls._meta.required_fields and any(k not in fields for k
-                                             in cls._meta.required_fields):
-            message = ("The '{0}' object cannot be created because it must "
-                       "contain all of the required fields: {1}")
-            message = message.format(cls.__name__,
-                                     ', '.join(cls._meta.required_fields))
-            raise TypeError(message)
-
         new = cls(**fields)
         new.save(safe=safe)
 
@@ -260,63 +296,55 @@ class Model(object):
                             "'{1}' attribute has not been set.".format(
                                 self.__class__.__name__, '_id'))
 
-        safe = safe or self._meta.safe
-        self._meta.db.remove({'_id': id}, safe=safe)
+        kwargs = {'safe': safe}
+        _set_write_concern(kwargs, self._meta.safe)
+
+        self._meta.db.remove({'_id': id}, **kwargs)
 
         self._document = {}
 
     @classmethod
-    def find(cls, *qs, **fields):
+    def find(cls, q=None, *qs, **fields):
         """Gets multiple documents from the database.
 
         This will find a return multiple documents matching the query
         specified through ``**fields``. If ``sort`` has been defined on
         the ``Meta`` class it will be used to order the records.
 
-        :param \*qs: :class:`~simon.query.Q` objects to use with the
-                      query.
+        :param q: (optional) A logical query to use with the query.
+        :type q: :class:`~simon.query.Q`.
+        :param \*qs: **DEPRECATED** Use ``q`` instead.
         :type \*qs: \*args.
         :param \*\*fields: Keyword arguments specifying the query.
         :type \*\*fields: \*\*kwargs.
         :returns: :class:`~simon.base.QuerySet` -- query set containing
                   objects matching ``query``.
 
+        .. versionchanged:: 0.3.0
+           Deprecating ``qs`` in favor of ``q``
+
         .. versionadded:: 0.1.0
         """
 
-        # Add all Q objects to the filter
-        for q in qs:
-            if isinstance(q, Q):
-                fields.update(q._filter)
+        if qs:
+            warnings.warn('qs has been deprecated. Please use q instead.',
+                          DeprecationWarning)
+            for filter in qs:
+                q._filter.update(filter._filter)
 
-        query = map_fields(cls, fields, flatten_keys=True,
-                           with_operators=True)
-
-        # If querying by the _id, make sure it's an Object ID
-        if '_id' in query:
-            query['_id'] = guarantee_object_id(query['_id'])
-
-        # Find all of the matching documents.
-        docs = cls._meta.db.find(query)
-
-        qs = QuerySet(docs, cls)
-
-        if cls._meta.sort:
-            # If the model has a default sort, apply it.
-            qs = qs.sort(*cls._meta.sort)
-
-        return qs
+        return cls._find(q=q, **fields)
 
     @classmethod
-    def get(cls, *qs, **fields):
+    def get(cls, q=None, *qs, **fields):
         """Gets a single document from the database.
 
         This will find and return a single document matching the
         query specified through ``**fields``. An exception will be
         raised if any number of documents other than one is found.
 
-        :param \*qs: :class:`~simon.query.Q` objects to use with the
-                     query.
+        :param q: (optional) A logical query to use with the query.
+        :type q: :class:`~simon.query.Q`.
+        :param \*qs: **DEPRECATED** Use ``q`` instead.
         :type \*qs: \*args.
         :param \*\*fields: Keyword arguments specifying the query.
         :type \*\*fields: \*\*kwargs.
@@ -324,43 +352,19 @@ class Model(object):
         :raises: :class:`~simon.Model.MultipleDocumentsFound`,
                  :class:`~simon.Model.NoDocumentFound`
 
+        .. versionchanged:: 0.3.0
+           Deprecating ``qs`` in favor of ``q``
+
         .. versionadded:: 0.1.0
         """
 
-        # Add all Q objects to the filter
-        for q in qs:
-            if isinstance(q, Q):
-                fields.update(q._filter)
+        if qs:
+            warnings.warn('qs has been deprecated. Please use q instead.',
+                          DeprecationWarning)
+            for filter in qs:
+                q._filter.update(filter._filter)
 
-        # Convert the field spec into a query by mapping any necessary
-        # fields.
-        query = map_fields(cls, fields, flatten_keys=True,
-                           with_operators=True)
-
-        # If querying by the _id, make sure it's an Object ID
-        if '_id' in query:
-            query['_id'] = guarantee_object_id(query['_id'])
-
-        # Find all of the matching documents. find_one() could be used
-        # here instead, but that would return the *first* matching
-        # document, not the *only* matching document. In order to know
-        # if a number of documents other than one was found, find()
-        # and count() must be used instead. Because MongoDB uses
-        # cursors, not data needs to be transferred until the result
-        # set is sliced later on.
-        docs = cls._meta.db.find(query)
-        count = docs.count()
-        if not count:
-            raise cls.NoDocumentFound(
-                "'{0}' matching query does not exist.".format(cls.__name__))
-        elif count > 1:
-            raise cls.MultipleDocumentsFound(
-                '`get()` returned more than one "{0}". It returned {1}!'
-                ' The document spec was: {2}'.format(
-                    cls.__name__, count, fields))
-
-        # Return an instantiated object for the retrieved document
-        return cls(**docs[0])
+        return cls._find(find_one=True, q=q, **fields)
 
     @classmethod
     def get_or_create(cls, safe=False, **fields):
@@ -427,12 +431,6 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
-        id = self._document.get('_id')
-        if not id:
-            raise TypeError("The '{0}' object cannot be updated because its "
-                            "'{1}' attribute has not been set.".format(
-                                self.__class__.__name__, '_id'))
-
         # There needs to be something to update
         if field is None and not fields:
             raise ValueError('No fields have been specified.')
@@ -447,32 +445,7 @@ class Model(object):
         for k, v in fields.items():
             update[k] = v
 
-        update = map_fields(self.__class__, update, flatten_keys=True)
-
-        safe = safe or self._meta.safe
-        self._meta.db.update({'_id': id}, {'$inc': update}, safe=safe)
-
-        # After updating the document in the database, the instance
-        # needs to be updated as well. Depending on the size of the
-        # document, it may be time consuming to reload the entire thing.
-        # Fortunately there is a way to just load the fields that have
-        # been updated. Build a dictionary containing the keys of the
-        # fields that need to be reloaded and retrieve them from the
-        # database. Then apply the new values to the instance
-        fields = dict((k, 1) for k in update.keys())
-        doc = self._meta.db.find_one({'_id': id}, fields)
-
-        # There's no need to update the _id
-        doc.pop('_id')
-
-        doc = update_nested_keys(self._document, doc)
-
-        for k, v in doc.iteritems():
-            # Normally I'd use self._document[k] = v, but if a value
-            # has already been associated with an attribute, assigning
-            # the new value through _document won't update the
-            # attribute's value.
-            setattr(self, k, v)
+        self._update({'$inc': update}, safe=safe)
 
     def raw_update(self, fields, safe=False):
         """Performs an update using a raw document.
@@ -503,31 +476,7 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
-        id = self._document.get('_id')
-        if not id:
-            raise TypeError("The '{0}' object cannot be updated because its "
-                            "'{1}' attribute has not been set.".format(
-                                self.__class__.__name__, '_id'))
-
-        safe = safe or self._meta.safe
-
-        self._meta.db.update({'_id': id}, fields, safe=safe)
-
-        # After saving the document, the entire document must be
-        # reloaded in order to update the instance.
-        doc = self._meta.db.find_one({'_id': self.id})
-
-        # There's no need to update the _id
-        doc.pop('_id')
-
-        doc = update_nested_keys(self._document, doc)
-
-        for k, v in doc.iteritems():
-            # Normally I'd use self._document[k] = v, but if a value
-            # has already been associated with an attribute, assigning
-            # the new value through _document won't update the
-            # attribute's value.
-            setattr(self, k, v)
+        self._update(fields, safe=safe)
 
     def remove_fields(self, fields, safe=False):
         """Removes the specified fields from the document.
@@ -557,20 +506,6 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
-        id = self._document.get('_id')
-        if not id:
-            raise TypeError("The '{0}' object cannot be updated because its "
-                            "'{1}' attribute has not been set.".format(
-                                self.__class__.__name__, '_id'))
-
-        if self._meta.required_fields and any(k in fields for k
-                                              in self._meta.required_fields):
-            message = ("The '{0}' object cannot be updated because it must "
-                       "contain all of the required fields: {1}")
-            message = message.format(self.__class__.__name__,
-                                     ', '.join(self._meta.required_fields))
-            raise TypeError(message)
-
         # fields can contain a single item as a string. If it's not a
         # list or tuple, make it one. Otherwise the generators below
         # would iterate over each character in the string rather than
@@ -578,26 +513,9 @@ class Model(object):
         if not isinstance(fields, (list, tuple)):
             fields = (fields,)
 
-        update = dict((k, 1) for k in fields)
-        update = map_fields(self.__class__, update, flatten_keys=True)
+        fields = dict((k, 1) for k in fields)
 
-        safe = safe or self._meta.safe
-        self._meta.db.update({'_id': id}, {'$unset': update}, safe=safe)
-
-        # The fields also need to be removed from the object
-        for k, v in update.items():
-            if '.' in k:
-                # If the key contains a ., it is pointing to an embedded
-                # document, so remove the nested key from the
-                # dictionary; there is no attribute to remove from the
-                # instance
-                self._document = remove_nested_key(self._document, k)
-            else:
-                try:
-                    delattr(self, k)
-                except AttributeError:
-                    # Silently handle attributes that don't exist
-                    pass
+        self._update({'$unset': fields}, safe=safe)
 
     def save(self, safe=False):
         """Saves the object to the database.
@@ -619,14 +537,6 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
-        if self._meta.required_fields and any(k not in self._document for k
-                                              in self._meta.required_fields):
-            message = ("The '{0}' object cannot be saved because it must "
-                       "contain all of the required fields: {1}")
-            message = message.format(self.__class__.__name__,
-                                     ', '.join(self._meta.required_fields))
-            raise TypeError(message)
-
         # Associate the current datetime (in UTC) with the created
         # and modified fields. While Python can store datetimes with
         # microseconds, BSON only supports milliseconds. Rather than
@@ -635,29 +545,17 @@ class Model(object):
         # instance.
         if self._meta.auto_timestamp:
             now = current_datetime()
-            if not ('_id' in self._document and 'created' in self):
-                self.created = now
-            self.modified = now
+            if not ('_id' in self._document or 'created' in self):
+                self._document['created'] = now
+            self._document['modified'] = now
 
-        # _id should never be overwritten. In order to do that, it's a
-        # good idea to pop it out of the document. Popping it out of
-        # the real document could lead to bad things happening. Instead,
-        # capture a reference to the document and pop it out of that.
-        doc = self._document.copy()
-        id = doc.pop('_id', None)
+        # Use a copy of the internal document so _id can be safely
+        # removed.
+        fields = self._document.copy()
+        fields.pop('_id', None)
 
-        doc = map_fields(self.__class__, doc)
-
-        safe = safe or self._meta.safe
-
-        if id:
-            # The document already exists, so it can just be updated.
-            id = guarantee_object_id(id)
-            self._meta.db.update({'_id': id}, doc, safe=safe)
-        else:
-            # When the document doesn't exist, insert() can be used to
-            # created it and get back the _id.
-            self._document['_id'] = self._meta.db.insert(doc, safe=safe)
+        # Use upsert=True so new documents can be inserted.
+        self._update(fields, upsert=True, safe=safe)
 
     def save_fields(self, fields, safe=False):
         """Saves only the specified fields.
@@ -690,13 +588,6 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
-        # Make sure the document has already been saved
-        id = self._document.get('_id')
-        if not id:
-            raise TypeError("The '{0}' object cannot be updated because its "
-                            "'{1}' attribute has not been set.".format(
-                                self.__class__.__name__, '_id'))
-
         # fields can contain a single item as a string. If it's not a
         # list or tuple, make it one. Otherwise the generators below
         # would iterate over each character in the string rather than
@@ -705,22 +596,10 @@ class Model(object):
             fields = (fields,)
 
         update = dict((k, 1) for k in fields)
-        update = map_fields(self.__class__, update, flatten_keys=True)
 
-        # Failing to make sure all of the fields exist before saving
-        # them would result in assigning blank values to keys. In some
-        # cases this could result in unexpected documents being returned
-        # in queries.
-        try:
-            update = dict((k, get_nested_key(self._document, k)) for
-                          k, v in update.items())
-        except KeyError:
-            raise AttributeError("The '{0}' object does not have all of the "
-                                 "specified fields.".format(
-                                     self.__class__.__name__))
-
-        safe = safe or self._meta.safe
-        self._meta.db.update({'_id': id}, {'$set': update}, safe=safe)
+        # Set use_internal so that the real values will come from
+        # self._document instead of the document passed to _update().
+        self._update({'$set': update}, use_internal=True, safe=safe)
 
     def update(self, safe=False, **fields):
         """Performs an atomic update.
@@ -748,31 +627,260 @@ class Model(object):
         .. versionadded:: 0.1.0
         """
 
+        self._update({'$set': fields}, safe=safe)
+
+    # Database interaction methods
+
+    @classmethod
+    def _find(cls, q=None, find_one=False, **fields):
+        """Find documents in the database.
+
+        This method will find documents in the database matching the
+        specified query. When ``find_one`` is set to ``False``, a
+        :class:`~simon.query.QuerySet` will be returned containing all
+        of the matching documents. When it is set to ``True``, the
+        single matching document will be returned as an instance of the
+        model.
+
+        When querying with ``find_one`` set to ``True``,
+        ``NoDocumentFound`` will be raised when no documents match the
+        query, ``MultipleDocumentsFound`` will be raised when more than
+        one matches.
+
+        :param q: (optional) A logical query to use with the query.
+        :type q: :class:`~simon.query.Q`.
+        :param find_one: Whether or not the query should only return one
+                         document.
+        :type find_one: bool.
+        :param \*\*fields: Keyword arguments specifying the query.
+        :type \*\*fields: \*\*kwargs.
+        :returns: :class:`~simon.base.Model` or
+                  :class:`~simon.base.QuerySet` -- the document or query
+                  set containing objects matching ``query``.
+        :raises: :class:`~simon.Model.MultipleDocumentsFound`,
+                 :class:`~simon.Model.NoDocumentFound`
+
+        .. versionadded:: 0.3.0
+        """
+
+        # If there is a Q object, add it to the spec document.
+        if isinstance(q, Q):
+            fields.update(q._filter)
+
+        query = map_fields(cls, fields, flatten_keys=True, with_operators=True)
+
+        # If querying by the _id, make sure it's an Object ID.
+        if '_id' in query:
+            query['_id'] = guarantee_object_id(query['_id'])
+
+        # Find all of the matching documents.
+        docs = cls._meta.db.find(query)
+
+        if find_one:
+            count = docs.count()
+
+            exception = None
+            if not count:
+                exception = cls.NoDocumentFound
+                message = "'{0}' matching query does not exist."
+                message = message.format(cls.__name__)
+            elif count > 1:
+                exception = cls.MultipleDocumentsFound
+                message = ("The query returned more than one '{0}'. It "
+                           "returned {1}! The document spec was: {2}.")
+                message = message.format(cls.__name__, count, fields)
+
+            if exception:
+                raise exception(message)
+
+            result = cls(**docs[0])
+        else:
+            result = QuerySet(docs, cls)
+
+            if cls._meta.sort:
+                # Apply the default sort for the model.
+                result = result.sort(*cls._meta.sort)
+
+        return result
+
+    def _update(self, fields, upsert=False, use_internal=False, **kwargs):
+        """Update documents in the database.
+
+        There are a few actions that need to be done in conjunction with
+        performing an update. Other methods can call this one to ensure
+        that:
+
+        - ``_id`` is checked (can be overridden by ``upsert``)
+        - ``Meta.required_fields`` is enforced
+        - field names are mapped according to ``Meta.field_map``
+        - ``w`` and ``safe`` are correctly handled
+        - atomic updates that go directly to the database also update
+          the instance
+
+        This method currently supports the following optional settings:
+
+        ======== =======================================================
+        Name     Description
+        ======== =======================================================
+        ``safe`` (optional) **DEPRECATED** use ``w`` instead
+        ``w``    (optional) int -- the number of servers in a replica
+                 set that must receive the update for it to be
+                 successful
+        ======== =======================================================
+
+        :param fields: The update to perform.
+        :type fields: dict.
+        :param upsert: (optional) Whether or not to insert new
+                       documents.
+        :type upsert: bool.
+        :param use_internal: (optional) Whether or not to use the values
+                             in the internal document.
+        :type use_internal: bool.
+        :param \*\*kwargs: The optional settings.
+        :type \*\*kwargs: \*\*kwargs.
+        :raises: :class:`TypeError`
+
+        .. versionadded:: 0.3.0
+        """
+
+        # Save characters
+        cls = self.__class__
+
+        def map_field_names_and_values(fields):
+            """Maps field names and values.
+
+            This method will take care of mapping the keys of ``fields``
+            to the correct syntax for the query.
+
+            When the  ``use_internal`` argument of
+            :meth:`~simon.Model._update()` is set to ``True``, it will
+            also update the values of ``fields`` from the internal
+            document. If a field does not exist in the internal
+            document, :class:`AttributeError` will be raised.
+
+            :param fields: The document containing keys to map.
+            :type fields: dict.
+            :raises: :class:`AttributeError`
+
+            .. versionadded:: 0.3.0
+            """
+
+            fields = map_fields(cls, fields, flatten_keys=True)
+            if use_internal:
+                try:
+                    fields = dict((k, get_nested_key(self._document, k))
+                                  for k in fields.keys())
+                except KeyError:
+                    # KeyError will be raised by get_nested_key() when a
+                    # field isn't part of the internal document.
+                    message = ("The '{0}' object does not have all of the "
+                               "specified fields.".format(cls.__name__))
+                    raise AttributeError(message)
+            return fields
+
+        # _id is required for updates unless the upsert flag has been
+        # set by the caller.
+        if not (upsert or '_id' in self._document):
+            message = ("The '{0}' object cannot be updated because its '_id' "
+                       "attribute has not been set.".format(cls.__name__))
+            raise TypeError(message)
+
+        # Capture the _id and make sure it's an Object ID
         id = self._document.get('_id')
+        if id:
+            id = guarantee_object_id(id)
+            # Updates (as opposed to inserts) need spec to match against
+            kwargs['spec'] = {'_id': id}
+
+        # Map all the field names and values
+        if is_atomic(fields):
+            for k, v in fields.iteritems():
+                fields[k] = map_field_names_and_values(v)
+        else:
+            fields = map_field_names_and_values(fields)
+        # When placing fields in kwargs, make a copy so changes to
+        # fields don't affect kwargs
+        kwargs['document'] = fields.copy()
+
+        # Enforce the required fields
+        if self._meta.required_fields:
+            has_required_fields = True
+
+            if is_atomic(fields):
+                # If performing an update that would result in fields
+                # being removed, make sure that none of the required
+                # fields are being removed.
+                if '$unset' in fields:
+                    if any(k in self._meta.required_fields for k
+                            in fields['$unset']):
+                        has_required_fields = False
+            else:
+                # For a document replacement, all of the required fields
+                # must appear in the update.
+                if any(k not in fields for k in self._meta.required_fields):
+                    has_required_fields = False
+
+            if not has_required_fields:
+                message = ("The '{0}' object cannot be updated because it "
+                           "must contain all of the required fields: {1}.")
+                message = message.format(cls.__name__,
+                                         ', '.join(self._meta.required_fields))
+                raise TypeError(message)
+
+        # Handling the write concern argument has been pushed off to
+        # another method that is aware of what PyMongo supports.
+        _set_write_concern(kwargs, self._meta.safe)
+
+        # Which function are we calling?
         if not id:
-            raise TypeError("The '{0}' object cannot be updated because its "
-                            "'{1}' attribute has not been set.".format(
-                                self.__class__.__name__, '_id'))
+            f = self._meta.db.insert
+        else:
+            f = self._meta.db.update
 
-        update = map_fields(self.__class__, fields, flatten_keys=True)
+        result = f(**kwargs)
 
-        safe = safe or self._meta.safe
-        self._meta.db.update({'_id': id}, {'$set': update}, safe=safe)
+        if not id:
+            # insert() will return the _id
+            self._document['_id'] = result
 
-        # After updating the document in the database, the instance
-        # needs to be updated as well. Depending on the size of the
-        # document, it may be time consuming to reload the entire thing.
-        # Fortunately there is a way to just load the fields that have
-        # been updated. Build a dictionary containing the keys of the
-        # fields that need to be reloaded and retrieve them from the
-        # database. Then apply the new values to the instance
-        fields = dict((k, 1) for k in update.keys())
-        doc = self._meta.db.find_one({'_id': id}, fields)
+        # For atomic updates, make sure the updates find their way back
+        # to the internal document.
+        # Right now this happens for all atomic updates, including when
+        # called from save_fields(). I'd like to turn that one off.
+        if is_atomic(fields):
+            # After updating the document in the database, the instance
+            # needs to be updated as well. Depending on the size of the
+            # document, it may be time consuming to reload the entire
+            # thing. Fortunately there is a way to just load the fields
+            # that have been updated. Build a dictionary containing the
+            # keys of the fields that need to be reloaded and retrieve
+            # them from the database. Then apply the new values to the
+            # instance's interal document.
 
-        doc = update_nested_keys(self._document, doc)
+            unset = fields.pop('$unset', None)
+            if unset:
+                # The fields also need to be removed from the object
+                for k, v in unset.items():
+                    if '.' in k:
+                        # If the key contains a ., it is pointing to an
+                        # embedded document, so remove the nested key
+                        # from the dictionary; there is no attribute to
+                        # remove from the instance.
+                        self._document = remove_nested_key(self._document, k)
+                    else:
+                        self._document.pop(k, None)
 
-        for k, v in doc.items():
-            setattr(self, k, v)
+            # If the only operation was an $unset, we're done.
+            if not fields:
+                return
+
+            fields = dict((k, 1) for v in fields.values() for k in v.keys())
+            doc = self._meta.db.find_one({'_id': id}, fields)
+
+            # There's no need to update the _id
+            doc.pop('_id', None)
+
+            self._document = update_nested_keys(self._document, doc)
 
     # String representation methods
 
@@ -818,7 +926,7 @@ class Model(object):
             if '__' in name:
                 # map_fields() requires a dict, so make a simple one and
                 # then capture the first (only) key in the resulting
-                # dict
+                # dict.
                 mapped_name = map_fields(self.__class__, {name: 1},
                                          flatten_keys=True)
                 mapped_name = mapped_name.keys()[0]
@@ -845,13 +953,30 @@ class Model(object):
             raise AttributeError(
                 "The '{0}' attribute cannot be overwritten.".format(name))
 
-        # Set the attribute on the object. Trying to do this with a
-        # simple assignment would result in recursion error.
-        object.__setattr__(self, name, value)
+        if name in self._meta.core_attributes:
+            # Set the attribute on the object. Trying to do this with a
+            # simple assignment would result in recursion error.
+            object.__setattr__(self, name, value)
+        else:
+            # Set the attribute in the internal document.
 
-        if name not in self._meta.core_attributes:
-            name = self._meta.field_map.get(name, name)
-            self._document[name] = value
+            # map_fields() requires a dict, so make a simple one and
+            # then capture the first (only) key in the resulting
+            # dict.
+            mapped_name = map_fields(self.__class__, {name: 1},
+                                     flatten_keys=True)
+            mapped_name = mapped_name.keys()[0]
+
+            # Build a dictionary that can be applied to the internal
+            # document dictionary with update_nested_keys().  Do this
+            # by iterating through the fields in mapped_name from right
+            # to left.
+            keys = mapped_name.split('.')
+            keys.reverse()
+            for x in keys:
+                value = {x: value}
+
+            self._document = update_nested_keys(self._document, value)
 
     # Rich comparison methods
 
