@@ -4,9 +4,11 @@ from pymongo import uri_parser
 try:
     # pymongo 2.4+
     from pymongo import MongoClient, MongoReplicaSetClient
+    pymongo_supports_mongoclient = True
 except ImportError:
     from pymongo import (Connection as MongoClient,
                          ReplicaSetConnection as MongoReplicaSetClient)
+    pymongo_supports_mongoclient = False
 
 from .exceptions import ConnectionError
 
@@ -42,6 +44,7 @@ def connect(host='localhost', name=None, username=None, password=None,
     .. versionchanged:: 0.2.0
        ``connect()`` now accepts ``replica_set`` as a kwarg, it is
        preferred over ``replicaSet``
+
     .. versionadded:: 0.1.0
     """
 
